// ============================================================================
//  Project:	Course "Object-Oriented Programming (in Java)"
//  Version:	Winter 2014/15, University of Halle
//  Module:	Maze.java
//  Purpose:	Homework 11: Small Computer Game (Rogue/Nethack/Larn-Style)
//  LastChange:	23.01.2015
//  Language:	Java
//  Authors:	Stefan Brass
//  Email:	brass@informatik.uni-halle.de
//  Address:	Universitaet Halle, Inst. f. Informatik, D-06099 Halle, Germany
//  Copying:	You may do what you want with this, but don't make me
//		responsible. This code probably contains bugs.
// ============================================================================

import java.awt.*;
import java.awt.event.*;

// ============================================================================
// Interface zur Ausgabe von Texten (z.B. TextArea im Fenster):
// ============================================================================

interface MazeOutput {
	void append(String text);
	void clear();
}

// ============================================================================
// Maze - Eigentliche Spiel-Logik (Ausfuehrung von Zuegen etc.)
// ============================================================================

class Maze {

	// Konstante fuer Daten der Karte:
	private static final String[] MAP_DATA = {
		"########################################",
		"###################.####################",
		"###.....#####.......#######........#####",
		"###.....#####.#####.#######.######.#####",
		"###.....#####.#####.#######.######.#####",
		"#######.#####.#####.........######.#####",
		"#######.#####.#############.######.#####",
		"#######.......#####.......#.######.#####",
		"#######.###########.#####.#.######.#####",
		"#######.###########.#..<#.#.######.#####",
		"#######........@###.#.###.#......#.#####",
		"#######.........###.#.....#.############",
		"#######.........###.#######.############",
		"#######.........###.........############",
		"########################################",
		"########################################",
	};

	// Attribut fuer Karten-Objekt:
	private Map map;

	// Attribut fuer Spieler-Objekt:
	private Player player;

	// Attribute fuer Feldgroesse, Start- und Zielposition:
	private int sizeX;
	private int sizeY;
	private int startX;
	private int startY;
	private int goalX;
	private int goalY;

	// Attribut fuer Objekt zur Ausgabe von Texten:
	private MazeOutput output;

	// Anzahl Zuege bisher:
	private int numMoves;

	// Spiel gewonnen/verloren:
	private boolean gameWon;
	private boolean gameLost;

	// Hilfsfunktion zum Anlegen des Spieler- und des Karten-Objektes:
	private void init(String[] map_data) {
		// Y-Groesse des Spielfeldes: Anzahl Strings im Array:
		this.sizeY = map_data.length;

		// X-Groesse des Spielfeldes: Maximal String-Laenge im Array:
		this.sizeX = 1;
		for(int y = 0; y < this.sizeY; y++)
			if(map_data[y].length() > this.sizeX)
				this.sizeX = map_data[y].length();

		// Karten-Objekt anlegen:
		this.map = new Map(this.sizeX, this.sizeY);

		// Start und Ziel ungueltig initialisieren:
		this.startX = -1;
		this.startY = -1;
		this.goalX = -1;
		this.goalY = -1;

		// Einzelne Spielfelder anlegen:
		for(int x = 0; x < this.sizeX; x++) {
			for(int y = 0; y < this.sizeY; y++) {
				if(x >= map_data[y].length())
					this.map.makeField(x, y, '#');
				else {
					char c = map_data[y].charAt(x);
					this.map.makeField(x, y, c);
					if(c == '@') {
						this.startX = x;
						this.startY = y;
					}
					if(c == '<') {
						this.goalX = x;
						this.goalY = y;
					}
				}
			}
		}

		// Fehler, falls startX/startY nicht gesetzt:
		if(this.startX < 0 || this.startY < 0) {
			System.err.println("@ fehlt in Daten der Karte");
			System.exit(1);
		}

		// Fehler, falls goalX/goalY nicht gesetzt:
		if(this.goalX < 0 || this.goalY < 0) {
			System.err.println("< fehlt in Daten der Karte");
			System.exit(1);
		}

		// Spieler-Objekt anlegen:
		this.player = new Player(this.startX, this.startY);

		// Bisher keine Zuege, Spiel laeuft noch:
		this.numMoves = 0;
		this.gameWon = false;
		this.gameLost = false;
	}

	// Konstruktor:
	Maze() {
		// Ausgabeobjekt wird erst spaeter gesetzt:
		this.output = null;

		// Kartendaten verarbeiten, Karten- und Spieler-Objekt anlegen:
		init(MAP_DATA);
	}

	// Ausgabeobjekt setzen:
	void setOutput(MazeOutput output) {
		this.output = output;
	}

	// Statuszeile liefern:
	String getStatus() {
		// Spezialfall: Gewonnen/Verloren:
		if(this.gameWon) {
			return "Gewonnen nach " + this.numMoves +
				" Z\u00FCgen!";
		}
		if(this.gameLost) {
			return "Verloren nach " + this.numMoves +
				" Z\u00FCgen!";
		}

		// Normalfall: Zuege und Gesundheitspunkte anzeigen:
		return "Z\u00FCge: " + this.numMoves +
				"  HP: " + this.player.getHP();
	}

	// Hilfsmethode zur Ausgabe:
	void println(String s) {
		// Wenn GUI-Objekt nicht gesetzt, einfache Konsolen-Ausgabe:
		if(output == null)
			System.out.println(s);
		else
			output.append(s + "\n");
	}

	// Neustart des Spiels:
	void restart() {
		this.output.clear();
		init(MAP_DATA);
	}

	// Spielzug: Bewegung nach Norden (hoch):
	void goNorth() {
		go(0, -1);
	}

	// Spielzug: Bewegung nach Osten (rechts):
	void goEast() {
		go(1, 0);
	}

	// Spielzug: Bewegung nach Sueden (unten):
	void goSouth() {
		go(0, 1);
	}

	// Spielzug: Bewegung nach Westen (links):
	void goWest() {
		go(-1, 0);
	}

	// Ausfuehrung des Spielzugs (Bewegung):
	void go(int stepX, int stepY) {
		// Ausgabefeld loeschen:
		this.output.clear();

		// Laeuft das Spiel noch?
		if(this.gameWon) {
			println("Das Spiel ist gewonnen. " +
				"Bitte Beenden oder Neustart.");
			return;
		}
		if(this.gameLost) {
			println("Das Spiel ist verloren. " +
				"Bitte Beenden oder Neustart.");
			return;
		}

		// Jetzt zaehlt es als Zug:
		this.numMoves++;

		// Aktuelle Position des Spielers:
		int playerX = this.player.getX();
		int playerY = this.player.getY();
		Field currField = this.map.getField(playerX, playerY);

		// Kann man das aktuelle Feld verlassen?
		String cantLeave = currField.cantLeave(this.player);
		if(cantLeave != null) {
			println(cantLeave);
			return;
		}

		// Grenzen des Spielfeldes pruefen:
		int nextX = playerX + stepX;
		int nextY = playerY + stepY;
		if(nextX < 0 || nextX >= this.sizeX ||
				nextY < 0 || nextY >= sizeY) {
			// Die Karte sollte so gemacht sein,
			// dass das nicht vorkommt.
			println("Da ist nur g\u00e4hnende Leere - das Nichts");
			return;
		}

		// Kann man das gewuenschte Feld betreten?
		Field nextField = this.map.getField(nextX, nextY);
		String cantEnter = nextField.cantEnter(this.player);
		if(cantEnter != null) {
			println(cantEnter);
			return;
		}

		// Nun Aktion durchfuehren:
		String leaveMessage = currField.leave(this.player);
		if(leaveMessage != null)
			println(leaveMessage);
		this.player.setXY(nextX, nextY);
		String enterMessage = nextField.enter(this.player);
		if(enterMessage != null)
			println(enterMessage);

		// Lebt der Spieler noch?
		if(this.player.getHP() <= 0) {
			println("Leider sind die Gesundheitspunkte " +
					"verbraucht. Game over!");
			this.gameLost = true;
			return;
		}

		// Position des Spielers neu holen, wurde ggf. teleportiert:
		playerX = this.player.getX();
		playerY = this.player.getY();

		// Ist der Spieler im Ziel?
		if(playerX == this.goalX && playerY == this.goalY) {
			println("Herzlichen Gl\u00FCckwunsch! " +
				"Sie haben den Ausgang gefunden.");
			this.gameWon = true;
		}
	}

	// Tastatur-Kommandos behandeln:
	void cmd(char c) {
		switch(c) {
			// Bewegung nach Norden (hoch):
			case 'k':
			case 'w':
				goNorth();
				break;

			// Bewegung nach Osten (rechts):
			case 'l':
			case 'd':
				goEast();
				break;

			// Bewegung nach Sueden (unten):
			case 'j':
			case 's':
				goSouth();
				break;

			// Bewegung nach Westen (links):
			case 'h':
			case 'a':
				goWest();
				break;

			// Falsche Taste (unbekannt):
			default:
				println("Unbekannte Taste: '" + c + "'");
		}
	}

	// Spieler-Position liefern, X-Koordinate:
	int getPlayerX() {
		return this.player.getX();
	}

	// Spieler-Position liefern, Y-Koordinate:
	int getPlayerY() {
		return this.player.getY();
	}

	// Ziel liefern, X-Koordinate:
	int getGoalX() {
		return this.goalX;
	}

	// Ziel liefern, Y-Koordinate:
	int getGoalY() {
		return this.goalY;
	}

	// Groesse der Karte liefern, X-Koordinate:
	int getSizeX() {
		return this.sizeX;
	}

	// Groesse der Karte liefern, Y-Koordinate:
	int getSizeY() {
		return this.sizeY;
	}

	// Zugriff auf ein Feld der Karte:
	Field getField(int x, int y) {
		return this.map.getField(x, y);
	}
}

// ============================================================================
// MazeGUI - Fenster fuer das Spiel (GUI)
// ============================================================================

class MazeGUI extends Frame implements WindowListener, ActionListener,
					KeyListener {

	// Attribute fuer Dialog-Elemente:
	private Label statusLine;
	private MapGUI mapCanvas;
	private MazeTextArea output;

	// Attribute fuer Menu-Eintraege:
	private MenuItem menuRestart;
	private MenuItem menuQuit;
	private MenuItem menuNorth;
	private MenuItem menuEast;
	private MenuItem menuSouth;
	private MenuItem menuWest;

	// Objekt zur Kontrolle des Spiels:
	Maze maze;

	// Hilfsfunktion zum Erzeugen eines GridBagConstraints-Objektes
	// (fuer den GidBagLayout Layout-Manager):
	private GridBagConstraints stdCell(int row, int col) {
		GridBagConstraints gbc = new GridBagConstraints();
		// Position im Gitter:
		gbc.gridx = col;
		gbc.gridy = row;
		// Hoehe und Breite ist eine Gitterzelle:
		gbc.gridwidth = 1;
		gbc.gridheight = 1;
		// Diese Zellen sind nur in der Breite flexibel:
		gbc.weightx = 100;
		gbc.weighty = 0;
		gbc.anchor = GridBagConstraints.WEST;
		gbc.fill = GridBagConstraints.HORIZONTAL;
		// Abstaende:
		gbc.insets = new Insets(4, 4, 4, 4);
		gbc.ipadx = 0;
		gbc.ipady = 0;
		return gbc;
	}

	// Version fuer breite Gitterzellen (ueber 2 Spalten, Hoehe fexibel):
	// (Methode wird hier im Moment nicht benoetigt)
	private GridBagConstraints wideCell(int row) {
		GridBagConstraints gbc = stdCell(row, 0);
		// Breite: 2 Gitterzellen:
		gbc.gridwidth = 2;
		// Flexibel auch in der Hoehe:
		gbc.weighty = 100;
		gbc.fill = GridBagConstraints.BOTH;
		return gbc;
	}

	// Und eine Spezialversion fuer den Knopf (feste Groesse, zentriert):
	// (Methode wird hier im Moment nicht benoetigt)
	private GridBagConstraints buttonCell(int row) {
		GridBagConstraints gbc = wideCell(row);
		// Gar nicht felxibel - feste Groesse:
		gbc.weightx = 0;
		gbc.weighty = 0;
		gbc.fill = GridBagConstraints.NONE;
		gbc.anchor = GridBagConstraints.CENTER;
		// Etwas mehr innerer Abstand:
		gbc.ipadx = 4;
		gbc.ipady = 4;
		return gbc;
	}

	// Menu-Eintrag erzeugen, Variante mit Shortcut:
	private MenuItem genMenuItem(Menu m, String Name, int shortcut) {
		MenuItem mi = new MenuItem(Name, new MenuShortcut(shortcut));
		mi.addActionListener(this);
		m.add(mi);
		return mi;
	}

	// Menu-Eintrag erzeugen, Variante ohne Shortcut:
	// (Methode wird hier im Moment nicht benoetigt)
	private MenuItem genMenuItem(Menu m, String Name) {
		MenuItem mi = new MenuItem(Name);
		mi.addActionListener(this);
		m.add(mi);
		return mi;
	}

	// Menu erzeugen:
	private MenuBar genMenu() {
		// Menu-Leiste erzeugen:
		MenuBar mb = new MenuBar();

		// Datei-Menu:
		Menu file = new Menu("Datei");
		this.menuRestart = genMenuItem(file, "Neustart");
		file.addSeparator();
		this.menuQuit = genMenuItem(file, "Ende/Quit", KeyEvent.VK_Q);
		mb.add(file);

		// Bewegungs-Menu:
		Menu move = new Menu("Bewegung");
		this.menuNorth = genMenuItem(move, "Norden", KeyEvent.VK_K);
		this.menuEast  = genMenuItem(move, "Osten", KeyEvent.VK_L);
		this.menuSouth = genMenuItem(move, "S\u00FCden",KeyEvent.VK_J);
		this.menuWest  = genMenuItem(move, "Westen", KeyEvent.VK_H);
		mb.add(move);

		// Menuleiste wird geliefert (fuer setMenuBar im Konstruktor):
		return mb;
	}

	// Konstruktor initialisiert Objekt fuer Fenster:
	public MazeGUI() {
		super("Maze");
		addWindowListener(this);
		addKeyListener(this);
		setBackground(Color.lightGray);

		// Spiel initialisieren:
		this.maze = new Maze();

		// Menu anlegen:
		setMenuBar(genMenu());

		// Layout-Manager:
		GridBagLayout layout = new GridBagLayout();
		setLayout(layout);

		// "Label" fuer Zusammenfassung (spaeter Gold, HitPoints, etc.)
		statusLine = new Label("Willkommen!");
		layout.setConstraints(statusLine, stdCell(0, 0));
			// d.h. oben (1. Zeile, im Moment gibt es nur Spalte 0)
		add(statusLine);

		// Die eigentliche Karte (von Canvas abgeleitete Klasse):
		mapCanvas = new MapGUI(this, maze);
		layout.setConstraints(mapCanvas, stdCell(1, 0));
			// d.h. Mitte (2. Zeile)
		add(mapCanvas);

		// "TextArea" fuer die Ausgabe (3 Zeilen):
		output = new MazeTextArea(this);
		layout.setConstraints(output, stdCell(2, 0));
			// d.h. unten (3. Zeile)
		add(output);
		maze.setOutput(output);

		// Focus beanspruchen:
		// requestFocus();

		// Fenstergroesse berechnen und anzeigen:
		pack();
		setVisible(true);
	}

	// Hilfsmethode zum Aktualisieren der Darstellung:
	private void updateWindow() {
		mapCanvas.repaint();
		statusLine.setText(maze.getStatus());
		//statusLine.repaint();
	}

	// Diese Fenster-Ereignisse werden ignoriert:
	public void windowActivated(WindowEvent e) {}
	public void windowDeactivated(WindowEvent e) {}
	public void windowIconified(WindowEvent e) {}
	public void windowDeiconified(WindowEvent e) {}
	public void windowOpened(WindowEvent e) {}

	// Wenn man das Fenster schliesst, wird es geloescht:
	public void windowClosing(WindowEvent e) {
		e.getWindow().setVisible(false);
		e.getWindow().dispose();
	}

	// Wenn das Fenster geloescht wird, endet das Programm:
	public void windowClosed(WindowEvent e) {
		System.exit(0);
	}

	// Wenn ein Menu-Punkt ausgewaehlt wird
	// (oder Enter im Textfeld eingegeben wird)
	// (oder ein Knopf gedrueckt wird - gibt es hier nicht),
	// wird diese Methode aufgerufen:
	public void actionPerformed(ActionEvent e) {
		if(e.getSource() == menuQuit) {
			this.setVisible(false);
			this.dispose();
			System.exit(0);
		}
		else if(e.getSource() == this.menuRestart)
			maze.restart();
		else if(e.getSource() == this.menuNorth)
			maze.goNorth();
		else if(e.getSource() == this.menuEast)
			maze.goEast();
		else if(e.getSource() == this.menuSouth)
			maze.goSouth();
		else if(e.getSource() == this.menuWest)
			maze.goWest();
		updateWindow();
	}

	// Taste gedrueckt:
	public void keyPressed(KeyEvent e) {
		int key = e.getKeyCode();
		if(key == KeyEvent.VK_LEFT || key == KeyEvent.VK_KP_LEFT)
			maze.goWest();
		else if(key == KeyEvent.VK_DOWN || key == KeyEvent.VK_KP_DOWN)
			maze.goSouth();
		else if(key == KeyEvent.VK_UP || key == KeyEvent.VK_KP_UP)
			maze.goNorth();
		else if(key == KeyEvent.VK_RIGHT ||
				key == KeyEvent.VK_KP_RIGHT)
			maze.goEast();
		updateWindow();
	}

	// Taste losgelassen:
	public void keyReleased(KeyEvent e) {
	}

	// Taste gedrueckt und losgelassen:
	public void keyTyped(KeyEvent e) {
		maze.cmd(e.getKeyChar());
		updateWindow();
	}
	
	// Hauptprogramm: Objekt dieser Klasse erzeugen:
	public static void main(String[] args) {
		new MazeGUI();
	}

}

// ============================================================================
// MazeTextArea - Textarea, aber mit Behandlung von Tastendruecken:
// ============================================================================

class MazeTextArea extends TextArea implements KeyListener, MazeOutput {

	// Attribut fuer Hauptfenster (zur Weitergabe von Key Events):
	private MazeGUI mainWindow;

	// Konstruktor:
	MazeTextArea(MazeGUI mainWindow) {
		super("", 3, 30, TextArea.SCROLLBARS_BOTH);
		setEditable(false);
		this.mainWindow = mainWindow;
		addKeyListener(this);
	}

	// Taste gedrueckt:
	public void keyPressed(KeyEvent e) {
		// Event an Hauptfenster weitergeben:
		this.mainWindow.keyPressed(e);
	}

	// Taste losgelassen:
	public void keyReleased(KeyEvent e) {
	}

	// Taste gedrueckt und losgelassen:
	public void keyTyped(KeyEvent e) {
		// Event an Hauptfenster weitergeben:
		this.mainWindow.keyTyped(e);
	}

	// Methode zum Loeschen des Inhalts:
	public void clear() {
		setText("");
	}
}

// ============================================================================
// MapGUI - Klasse zum Anzeigen einer Karte (Objekt der Benutzerschnittstelle):
// ============================================================================

class MapGUI extends Canvas implements KeyListener {
	// Konstanten fuer Groesse eines Feldes in Pixeln:
	private static final int FIELD_X = 16; // Groesse in X-Richtung (Width)
	private static final int FIELD_Y = 16; // Groesse in Y-Richtung (Height)

	// Objekt fuer Spielsteuerung:
	private Maze maze;

	// Hauptfenster (zur Weitergabe von Key Events):
	private MazeGUI mainWindow;

	// Konstruktor:
	MapGUI(MazeGUI mainWindow, Maze maze) {
		super();
		this.mainWindow = mainWindow;
		this.maze = maze;
		addKeyListener(this);
	}

	@Override public Dimension getPreferredSize() {
		int pixelX = this.maze.getSizeX() * FIELD_X;
		int pixelY = this.maze.getSizeY() * FIELD_Y;
		return new Dimension(pixelX, pixelY);
	}

	@Override public Dimension getMinimumSize() {
		int pixelX = this.maze.getSizeX() * FIELD_X;
		int pixelY = this.maze.getSizeY() * FIELD_Y;
		return new Dimension(pixelX, pixelY);
	}

	// Hilfsfunktion zum Zeichnen eines Zeichens:
	void drawChar(Graphics g, int x, int y, char c) {
		// Zeichenposition bestimmen:
		FontMetrics fm = g.getFontMetrics();
		int diffY = (FIELD_Y - fm.getHeight()) / 2 + fm.getAscent();
		int diffX = (FIELD_X - fm.charWidth('@')) / 2;

		// Zeichen in String umwandeln:
		String s = String.valueOf(c);
		g.drawString(s, x*FIELD_X + diffX, y*FIELD_Y + diffY);
	}

	// Um das Flackern zu vermeiden, wird der Hintergrund nicht geloescht:
	@Override public void update(Graphics g) {
		paint(g);
		// paint zeichnet sowieso die ganze Karte neu.
	}

	@Override public void paint(Graphics g) {
		// Groesse des Spielfeldes und Spielerposition bestimmen:
		int sizeX = this.maze.getSizeX();
		int sizeY = this.maze.getSizeY();
		int playerX = this.maze.getPlayerX();
		int playerY = this.maze.getPlayerY();
		int goalX = this.maze.getGoalX();
		int goalY = this.maze.getGoalY();


		// Schleife ueber allen Spielfeld-Positionen:
		for(int y = 0; y < sizeY; y++) {
			for(int x = 0; x < sizeX; x++) {
				Field f = this.maze.getField(x, y);

				// Alles in der direkten Umgebung ist sichtbar:
				boolean visible = 
					(playerX - 1 <= x &&
							x <= playerX + 1 &&
					 playerY - 1 <= y &&
							y <= playerY + 1);

				// Wenn es sichtbar ist, ist es spaetestens
				// jetzt bekannt:
				if(visible)
					f.makeKnown();

				// Hintergrund-Farbe waehlen:
				boolean isWall = f.isWall();
				if(!f.isKnown())
					g.setColor(Color.GRAY);
				else if(isWall && visible)
					g.setColor(Color.DARK_GRAY);
				else if(isWall && !visible)
					g.setColor(Color.BLACK);
				else if(!isWall && visible)
					g.setColor(Color.WHITE);
				else if(!isWall && !visible)
					g.setColor(Color.LIGHT_GRAY);
				else
					// This should never happen
					g.setColor(Color.RED);

				// Feld zeichnen:
				g.fillRect(x*FIELD_X, y*FIELD_Y,
						FIELD_X, FIELD_Y);

				// Ggf. Symbol in das Feld zeichnen:
				char sym = f.symbol();
				if(x == playerX && y == playerY) {
					g.setColor(Color.RED);
					drawChar(g, x, y, '@');
				}
				else if(visible && x == goalX && y == goalY) {
					g.setColor(Color.BLUE);
					drawChar(g, x, y, '<');
				}
				else if(f.isKnown() && sym != ' ' && sym != 0){
					if(f.isWall())
						g.setColor(Color.ORANGE);
					else
						g.setColor(Color.BLUE);
					drawChar(g, x, y, sym);
				}
			}
		}
	}

	// Taste gedrueckt:
	public void keyPressed(KeyEvent e) {
		// Event an Hauptfenster weitergeben:
		this.mainWindow.keyPressed(e);
	}

	// Taste losgelassen:
	public void keyReleased(KeyEvent e) {
	}

	// Taste gedrueckt und losgelassen:
	public void keyTyped(KeyEvent e) {
		// Event an Hauptfenster weitergeben:
		this.mainWindow.keyTyped(e);
	}
}

